use crate::{Context, Handler};
use axum::{
    extract::{Path, State},
    http::StatusCode,
    Json,
};
use serde::Deserialize;
use verax::{RevId, Status, Tag, TokenPayload};

#[derive(Deserialize)]
pub struct UpdateOperation {
    #[serde(default)]
    pub status: Option<Status>,
    #[serde(default)]
    pub tags: Option<Vec<Tag>>,
    pub memo: String,
    pub update_token: Option<TokenPayload>,
}

struct Update {
    id: RevId,
    operation: UpdateOperation,
}

#[async_trait::async_trait]
impl Handler for Update {
    type Ok = verax::Transaction;
    type Err = verax::Error;

    async fn handle(self, ledger: &Context) -> Result<Self::Ok, Self::Err> {
        let id = self.id;
        let memo = self.operation.memo;
        let update_token = self.operation.update_token;

        let id = if let Some(status) = self.operation.status {
            let transaction = ledger
                .ledger
                .change_status(id, status, memo.clone(), update_token.clone())
                .await?;
            transaction.revision_id
        } else {
            id
        };

        let id = if let Some(tags) = self.operation.tags {
            let transaction = ledger.ledger.set_tags(id, tags, memo, update_token).await?;
            transaction.revision_id
        } else {
            id
        };

        ledger
            .ledger
            .get_transactions(verax::Filter {
                revisions: vec![id],
                limit: 1,
                ..Default::default()
            })
            .await?
            .pop()
            .ok_or(verax::Error::TxNotFound)
    }
}

pub async fn handler(
    Path(info): Path<RevId>,
    State(ctx): State<Context>,
    Json(item): Json<UpdateOperation>,
) -> Result<Json<verax::Transaction>, (StatusCode, String)> {
    let update = Update {
        id: info,
        operation: item,
    };

    update
        .handle(&ctx)
        .await
        .map_err(|err| {
            (
                StatusCode::BAD_REQUEST,
                serde_json::to_string(&err).unwrap_or_default(),
            )
        })
        .map(Json)
}
