use crate::filter_expr::{opcode::OpCode, Value};
use std::collections::HashMap;

/// Executes operations that can be performed at compile time, and sets the initial_register to
/// the result. This is useful for optimizing the program, by executing at `compile` time as
/// much as possible, when the terms of the opcodes are not dependent on the runtime state.
pub fn calculate_static_values(mut opcodes: Vec<OpCode>) -> (Vec<OpCode>, bool) {
    let mut register = HashMap::new();
    let mut has_changed = false;

    opcodes.iter_mut().for_each(|opcode| {
        match &opcode {
            OpCode::LOAD(dst, value) => {
                register.insert(*dst, value.clone());
            }
            OpCode::JMP(_) | OpCode::HLT(_) | OpCode::LABEL(_) => {}
            OpCode::MOV(dst, src) => {
                if let Some(value) = register.remove(src) {
                    register.insert(*dst, value.clone());
                }
            }
            OpCode::LOAD_EXTERNAL(dst, _)
            | OpCode::AND(dst, _, _)
            | OpCode::OR(dst, _, _)
            | OpCode::NOT(dst, _)
            | OpCode::MOD(dst, _, _)
            | OpCode::NE(dst, _, _)
            | OpCode::GT(dst, _, _)
            | OpCode::LT(dst, _, _)
            | OpCode::NEG(dst, _)
            | OpCode::SHL(dst, _, _)
            | OpCode::XOR(dst, _, _)
            | OpCode::SHR(dst, _, _)
            | OpCode::GE(dst, _, _)
            | OpCode::LE(dst, _, _)
            | OpCode::CPY(dst, _) => {
                register.remove(dst);
            }
            OpCode::JEQ(reg, addr) => {
                if let Some(Value::Bool(true)) = register.get(reg) {
                    *opcode = OpCode::JMP(*addr);
                    has_changed = true;
                }
            }
            OpCode::JNE(reg, addr) => {
                if let Some(Value::Bool(false)) = register.get(reg) {
                    *opcode = OpCode::JMP(*addr);
                }
            }
            OpCode::EQ(dst, reg1, reg2) => {
                let value1 = if let Some(value) = register.get(reg1) {
                    value.clone()
                } else {
                    register.remove(dst);
                    return;
                };
                let value2 = if let Some(value) = register.get(reg2) {
                    value.clone()
                } else {
                    register.remove(dst);
                    return;
                };
                let result = value1 == value2;
                register.insert(*dst, result.into());
                *opcode = OpCode::LOAD(*dst, result.into());
            }
            OpCode::MUL(dst, reg1, reg2)
            | OpCode::SUB(dst, reg1, reg2)
            | OpCode::DIV(dst, reg1, reg2)
            | OpCode::ADD(dst, reg1, reg2) => {
                let number1 = match register.get(reg1) {
                    Some(Value::Number(number)) => *number,
                    _ => {
                        register.remove(dst);
                        return;
                    }
                };
                let number2 = match register.get(reg2) {
                    Some(Value::Number(number)) => *number,
                    _ => {
                        register.remove(dst);
                        return;
                    }
                };
                if let Some(calculated_value) = match opcode {
                    OpCode::ADD(_, _, _) => number1.checked_add(number2).map(Value::Number),
                    OpCode::MUL(_, _, _) => number1.checked_mul(number2).map(Value::Number),
                    OpCode::SUB(_, _, _) => number1.checked_sub(number2).map(Value::Number),
                    OpCode::DIV(_, _, _) => number1.checked_div(number2).map(Value::Number),
                    _ => None,
                } {
                    register.insert(*dst, calculated_value.clone());
                    *opcode = OpCode::LOAD(*dst, calculated_value);
                    has_changed = true;
                } else {
                    register.remove(dst);
                }
            }
        };
    });
    (opcodes, has_changed)
}
