use super::{
    compiler::Compiler,
    expr::{Expr, ExprOp, Variable},
    opcode::OpCode,
    parser::{parse_query, Query},
    runtime::execute,
    value::{Value, ValueOrRef},
    Addr, Error, Register,
};
use crate::Transaction;
use std::collections::HashMap;

#[derive(Debug, Clone)]
pub struct Filter<'a> {
    /// Query
    query: Query,
    /// List of external variables
    variables: Vec<Variable>,
    /// Unoptimized opcodes
    dbg_opcodes: Vec<OpCode>,
    /// Opcodes with human readable labels
    opcodes: Vec<OpCode>,
    /// The list of opcodes that make up the program, the labels has been converted into addresses
    opcodes_to_execute: Vec<OpCode>,
    /// The state of the register
    pub(crate) initial_register: HashMap<Register, ValueOrRef<'a>>,
    _phantom: std::marker::PhantomData<&'a ()>,
}

impl<'a> Filter<'a> {
    pub fn new(code: &str) -> Result<Self, Error> {
        let query = parse_query(code)?;
        let opcodes = query.where_clause.as_ref().map_or_else(
            || {
                Ok(vec![
                    OpCode::LOAD(0.into(), true.into()),
                    OpCode::HLT(0.into()),
                ])
            },
            |expr| Compiler::new(expr).compile(),
        )?;

        let instance = Self {
            query,
            dbg_opcodes: opcodes.clone(),
            opcodes: opcodes.clone(),
            variables: opcodes
                .iter()
                .filter_map(|x| match x {
                    OpCode::LOAD_EXTERNAL(_, name) => Some(name.clone()),
                    _ => None,
                })
                .collect(),
            opcodes_to_execute: Compiler::resolve_label_to_addr(opcodes)?,
            initial_register: HashMap::new(),
            _phantom: std::marker::PhantomData,
        };
        Ok(instance.optimize())
    }

    /// Optimizes the program
    pub fn optimize(mut self) -> Self {
        let mut new_opcodes = self.opcodes;
        loop {
            let (new_opcodes_, has_changed) = Compiler::optimize(new_opcodes);
            new_opcodes = new_opcodes_;
            if !has_changed {
                break;
            }
        }

        self.opcodes = Compiler::remove_dead_code(new_opcodes);
        self.initial_register.clear();
        self.opcodes_to_execute = Compiler::resolve_label_to_addr(
            self.opcodes
                .iter()
                .filter_map(|opcode| match opcode {
                    OpCode::LOAD(addr, value) => {
                        self.initial_register
                            .insert(*addr, ValueOrRef::Value(value.clone()));
                        None
                    }
                    _ => Some(opcode.clone()),
                })
                .collect::<Vec<_>>(),
        )
        .unwrap();
        self
    }

    /// Returns a human readable version of the compiled program (generated op-codes)
    pub fn debug(&self) -> String {
        self.dbg_opcodes
            .iter()
            .map(|x| match x {
                OpCode::HLT(_) | OpCode::LABEL(_) => x.to_string(),
                x => format!("\t{}", x.to_string()),
            })
            .collect::<Vec<_>>()
            .join("\n")
    }

    pub fn dump(&self) -> String {
        self.opcodes_to_execute
            .iter()
            .enumerate()
            .map(|(pos, opcode)| format!("{}: {}", pos, opcode.to_string()))
            .collect::<Vec<_>>()
            .join("\n")
    }

    pub fn execute(
        &'a self,
        external_variables: &'a HashMap<Variable, ValueOrRef<'a>>,
    ) -> Result<Value, Error> {
        execute(
            external_variables,
            &self.opcodes_to_execute,
            self.initial_register.clone(),
        )
    }
}
