//! Expression module

#![allow(warnings)]
use parser::Rule;
use std::num::ParseIntError;

mod compiler;
mod expr;
mod filter;
mod opcode;
mod parser;
mod runtime;
mod value;

#[cfg(test)]
mod tests;

pub use self::{filter::Filter, value::Value};

#[derive(
    Debug,
    Clone,
    Copy,
    PartialEq,
    Eq,
    Hash,
    derive_more::Display,
    derive_more::Deref,
    derive_more::DerefMut,
    derive_more::From,
    derive_more::Into,
)]
pub struct Register(usize);

#[derive(
    Debug,
    Clone,
    Copy,
    PartialEq,
    Eq,
    Hash,
    derive_more::Display,
    derive_more::Deref,
    derive_more::DerefMut,
    derive_more::From,
    derive_more::Into,
)]
pub struct Addr(usize);

impl Addr {
    pub fn next(&mut self) {
        self.0 += 1;
    }

    pub fn jump_to(&mut self, addr: &Addr) {
        self.0 = addr.0;
    }
}

#[derive(thiserror::Error, Debug, PartialEq)]
pub enum Error {
    #[error("Variable not found: {0}")]
    VariableNotFound(String),

    #[error("Program is out of bound")]
    OutOfBoundaries,

    #[error("Register {0} is out of boundaries")]
    RegisterOutOfBoundaries(Register),

    #[error("Empty registers: {0}")]
    EmptyRegister(Register),

    #[error("Value is not a number")]
    NotANumber,

    #[error("Math overflow")]
    Overflow,

    #[error("Unexpected rule: {0:?}")]
    UnexpectedRule(Rule),

    #[error("Parsing error: {0:?}")]
    ParsingError(#[from] pest::error::Error<Rule>),

    #[error("Internal parsing error: Missing next rule")]
    MissingNextRule,

    #[error("Error parsing number: {0}")]
    ParseIntError(#[from] ParseIntError),

    #[error("Invalid order: {0}")]
    InvalidOrder(String),

    #[error("Unknown label: {0}")]
    UnknownLabel(usize),

    #[error("Unexpected expression state")]
    UnexpectedExprState,
}
