#[macro_export]
/// Creates a wrapper for a hexahexadecimal identifier with a human prefix
macro_rules! BinaryId {
    ($id:ident, $suffix:expr) => {
        #[derive(
            Clone,
            Debug,
            Eq,
            PartialOrd,
            Ord,
            Hash,
            Default,
            PartialEq,
            borsh::BorshSerialize,
            borsh::BorshDeserialize,
        )]
        /// A unique identifier for $id
        pub struct $id {
            bytes: [u8; 32],
        }

        impl From<[u8; 32]> for $id {
            fn from(bytes: [u8; 32]) -> Self {
                Self { bytes }
            }
        }

        impl $id {
            /// Creates a new instance of $id from the raw bytes
            pub fn new(bytes: [u8; 32]) -> Self {
                Self { bytes }
            }

            /// Returns the raw bytes of the identifier
            pub fn as_bytes(&self) -> &[u8; 32] {
                &self.bytes
            }
        }

        impl std::str::FromStr for $id {
            type Err = $crate::id::Error;
            fn from_str(value: &str) -> Result<Self, Self::Err> {
                Ok(Self::try_from(value).unwrap_or_else(|_| {
                    use sha2::Digest;
                    let mut hasher = sha2::Sha256::new();
                    hasher.update(&value);
                    Self {
                        bytes: hasher.finalize().into(),
                    }
                }))
            }
        }

        impl Serialize for $id {
            fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
            where
                S: serde::Serializer,
            {
                serializer.collect_str(&self)
            }
        }

        impl<'de> Deserialize<'de> for $id {
            fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
            where
                D: serde::Deserializer<'de>,
            {
                use std::str::FromStr;
                let s = <String as serde::Deserialize>::deserialize(deserializer)?;
                // Use FromStr to parse the string and construct the struct
                $id::from_str(&s).map_err(serde::de::Error::custom)
            }
        }

        impl TryFrom<&str> for $id {
            type Error = $crate::id::Error;
            fn try_from(value: &str) -> Result<Self, Self::Error> {
                let (hrp, bytes) = bech32::decode(&value)?;
                let hrp = hrp.to_string();
                if hrp != $suffix {
                    return Err($crate::id::Error::InvalidPrefix(
                        stringify!($id).to_owned(),
                        $suffix.to_owned(),
                        hrp,
                    ));
                }

                bytes.try_into()
            }
        }

        impl TryFrom<&[u8]> for $id {
            type Error = $crate::id::Error;

            fn try_from(value: &[u8]) -> Result<Self, Self::Error> {
                if value.len() != 32 {
                    return Err($crate::id::Error::InvalidLength(
                        stringify!($id).to_owned(),
                        value.len(),
                        32,
                    ));
                }
                let mut bytes = [0u8; 32];
                bytes.copy_from_slice(&value);
                Ok(Self { bytes })
            }
        }

        impl TryFrom<Vec<u8>> for $id {
            type Error = $crate::id::Error;

            fn try_from(value: Vec<u8>) -> Result<Self, Self::Error> {
                if value.len() != 32 {
                    return Err($crate::id::Error::InvalidLength(
                        stringify!($id).to_owned(),
                        value.len(),
                        32,
                    ));
                }
                let mut bytes = [0u8; 32];
                bytes.copy_from_slice(&value);
                Ok(Self { bytes })
            }
        }

        impl std::fmt::Display for $id {
            fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
                let hrp = bech32::Hrp::parse($suffix).map_err(|_| std::fmt::Error)?;
                write!(
                    f,
                    "{}",
                    bech32::encode::<bech32::Bech32m>(hrp, &self.bytes)
                        .map_err(|_| std::fmt::Error)?
                )
            }
        }

        impl AsRef<[u8]> for $id {
            fn as_ref(&self) -> &[u8] {
                &self.bytes
            }
        }

        impl AsRef<[u8; 32]> for $id {
            fn as_ref(&self) -> &[u8; 32] {
                &self.bytes
            }
        }

        impl std::ops::Deref for $id {
            type Target = [u8; 32];
            fn deref(&self) -> &Self::Target {
                &self.bytes
            }
        }
    };
}
